"""
This handles `.ini` files, or more generally known as `config` files.
"""
import configparser
import re
from typing import Iterator
from typing import List
from typing import Tuple

from ..custom_types import NamedIO
from ..util.filetype import determine_file_type
from ..util.filetype import FileType
from .base import BaseTransformer
from .exceptions import ParsingError
from detect_secrets.filters.allowlist import get_allowlist_regexes


class ConfigFileTransformer(BaseTransformer):
    def should_parse_file(self, filename: str) -> bool:
        return True

    def parse_file(self, file: NamedIO) -> List[str]:
        try:
            return _parse_file(file)
        except configparser.Error:
            raise ParsingError


class EagerConfigFileTransformer(BaseTransformer):
    # NOTE: Currently eager, since `determine_file_type` is minimalistic right now.
    is_eager = True

    def should_parse_file(self, filename: str) -> bool:
        return determine_file_type(filename) == FileType.OTHER

    def parse_file(self, file: NamedIO) -> List[str]:
        try:
            return _parse_file(file, add_header=True)
        except configparser.Error:
            raise ParsingError


def _parse_file(file: NamedIO, add_header: bool = False) -> List[str]:
    """
    :raises: configparser.Error
    :raises: UnicodeDecodeError
    """
    lines: List[str] = []
    for key, value, line_number in IniFileParser(file, add_header=add_header):
        while len(lines) < line_number - 1:
            lines.append('')

        # Always add 'pragma: allowlist nextline secret' comments
        if _is_allowlist_nextline_secret_comment(value):
            lines.append(value)
            continue

        # We artificially add quotes here because we know they are strings
        # (because it's a config file), HighEntropyString will benefit from this,
        # and all other plugins don't care.
        if value[0] in {"'", '"'} and value[-1] == value[0]:
            # Strip out quotes, because we'll add our own.
            value = value[1:-1]

        value = value.replace('"', '\\"')
        lines.append(f'{key} = "{value}"')

    return lines


class EfficientParsingError(configparser.ParsingError):

    def append(self, lineno: int, line: str) -> None:
        """
        Rather than inefficiently add all the lines in the file
        to the error message like the CPython code from 1998,
        we just `return` because we will catch and `pass`
        the exception in `high_entropy_strings.py` anyway.
        """
        return


configparser.ParsingError = EfficientParsingError       # type: ignore


class IniFileParser:

    _comment_regex = re.compile(r'\s*[;#]')

    def __init__(self, file: NamedIO, add_header: bool = False) -> None:
        """
        :param add_header: whether or not to add a top-level [global] header.
        """
        self.parser = configparser.ConfigParser()
        self.parser.optionxform = str  # type: ignore

        content = file.read()
        if add_header:
            # This supports environment variables, or other files that look
            # like config files, without a section header.
            content = '[global]\n' + content

        self.parser.read_string(content)

        # Hacky way to keep track of line location
        file.seek(0)
        self.lines = [line.strip() for line in file.readlines()]
        self.line_offset = 0

    def __iter__(self) -> Iterator[Tuple[str, str, int]]:
        if not self.parser.sections():
            # To prevent cases where it's not an ini file, but the parser
            # helpfully attempts to parse everything to a DEFAULT section,
            # when not explicitly provided.
            raise configparser.Error

        for section_name in self.parser:
            for key, values in self.parser.items(section_name):
                for value, offset in self._get_value_and_line_offset(key, values):
                    if not value:
                        continue

                    yield key, value, offset

    def _get_value_and_line_offset(self, key: str, values: str) -> List[Tuple[str, int]]:
        """Returns the index of the location of key, value pair in lines.

        :param key: key, in config file.
        :param values: values for key, in config file. This is plural,
            because you can have multiple values per key. e.g.

            >>> key =
            ...     value1
            ...     value2
        """
        values_list = _construct_values_list(values)
        if not values_list:
            return []

        current_value_list_index = 0
        output = []

        for line_offset, line in enumerate(self.lines):
            # Check 'pragma: allowlist nextline secret' comment on a single line
            # The IniFileParser strips out comments however it is important to
            # persist this speific comment type so filtering works properly.
            if _is_allowlist_nextline_secret_comment(line):
                output.append((
                    line,
                    self.line_offset + line_offset + 1,
                ))
                continue

            # Check ignored lines before checking values, because
            # you can write comments *after* the value.
            if not line or self._comment_regex.match(line):
                continue

            # The first line is special because it's the only one with the variable name.
            # As such, we should handle it differently.
            if current_value_list_index == 0:
                # In situations where the first line does not have an associated value,
                # it will be an empty string. However, this regex still does its job because
                # it's not necessarily the case where the first line is a non-empty one.
                #
                # Therefore, we *only* advance the current_value_list_index when we identify
                # the key used.
                first_line_regex = re.compile(
                    r'^\s*{key}[ :=]+{value}'.format(
                        key=re.escape(key),
                        value=re.escape(values_list[current_value_list_index]),
                    ),
                )
                if first_line_regex.match(line):
                    output.append((
                        values_list[current_value_list_index],
                        self.line_offset + line_offset + 1,
                    ))
                    current_value_list_index += 1

                continue

            # There's no more values to iterate over.
            if current_value_list_index == len(values_list):
                if line_offset == 0:
                    line_offset = 1  # Don't want to count the same line again

                self.line_offset += line_offset
                self.lines = self.lines[line_offset:]

                break

            # This handles all other cases, when it isn't an empty or blank line.
            output.append((
                values_list[current_value_list_index],
                self.line_offset + line_offset + 1,
            ))
            current_value_list_index += 1
        else:
            self.lines = []

        return output


def _construct_values_list(values: str) -> List[str]:
    """
    This values_list is a strange construction, because of ini format.
    We need to extract the values with the following supported format:

        >>> key = value0
        ...     value1
        ...
        ...     # Comment line here
        ...     value2

    given that normally, either value0 is supplied, or (value1, value2),
    but still allowing for all three at once.

    Furthermore, with the configparser, we will get a list of values,
    and intermediate blank lines, but no comments. This means that we can't
    merely use the count of values' items to heuristically "skip ahead" lines,
    because we still have to manually parse through this.

    Therefore, we construct the values_list in the following fashion:
        1. Keep the first value (in the example, this is `value0`)
        2. For all other values, ignore blank lines.
    Then, we can parse through, and look for values only.
    """
    lines = values.splitlines()
    values_list = lines[:1]
    values_list.extend(filter(None, lines[1:]))
    return values_list


def _is_allowlist_nextline_secret_comment(line: str) -> bool:
    # Valid tuples for config file comments (start_char, end_char)
    comment_tuple = [('#', ''), (';', '')]

    for t in comment_tuple:
        if get_allowlist_regexes(comment_tuple=t, nextline=True).search(line):
            return True

    return False



"""
This code was extracted in part from
https://github.com/pre-commit/pre-commit-hooks. Using similar heuristic logic,
we adapted it to fit our plugin infrastructure, to create an organized,
concerted effort in detecting all type of secrets in code.

Copyright (c) 2014 pre-commit dev team: Anthony Sottile, Ken Struys

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""
import os
import re
from typing import Any
from typing import Generator
from typing import Optional
from typing import Set
from typing import Tuple

from ..core.potential_secret import PotentialSecret
from ..util.code_snippet import CodeSnippet
from .base import RegexBasedDetector


class PrivateKeyDetector(RegexBasedDetector):
    """
    Scans for private keys.

    This checks for private keys by determining whether the denylisted
    lines are present in the analyzed string.
    """

    secret_type = 'Private Key'
    MAX_FILE_SIZE: int = 8 * 1024

    begin_key_opening = r'(?P<begin_key>BEGIN'
    key_types = r'(?: DSA | EC | OPENSSH | PGP | RSA | SSH2 ENCRYPTED | )'
    begin_key_closing = r'PRIVATE KEY-*)'
    begin_key = fr'{begin_key_opening}{key_types}{begin_key_closing}'
    secret_key = r'(?P<secret_key>[A-Za-z0-9+\/\\\n]{10,}={0,3})'
    end_key = r'(?P<end_key>\n*-*END)?'

    denylist = [
        re.compile(
            r'{begin_key}{secret_key}{end_key}'.format(
                begin_key=begin_key,
                secret_key=secret_key,
                end_key=end_key,
            ),
        ),
        re.compile(r'PuTTY-User-Key-File-2:.{1,40}\n?Encryption:'),
    ]

    def __init__(self) -> None:
        self._analyzed_files: Set[str] = set()
        self._commit_hashes: Set[Tuple[str, str]] = set()

    def analyze_line(
            self,
            filename: str,
            line: str,
            line_number: int = 0,
            context: Optional[CodeSnippet] = None,
            raw_context: Optional[CodeSnippet] = None,
            commit_hash: Optional[str] = '',
            **kwargs: Any,
    ) -> Set[PotentialSecret]:
        output: Set[PotentialSecret] = set()

        output.update(
            super().analyze_line(
                filename=filename, line=line, line_number=line_number,
                context=context, raw_context=raw_context, **kwargs,
            ),
        )

        if output:
            return output

        # for git history
        if commit_hash:
            if (filename, commit_hash) not in self._commit_hashes:
                file_content = ''
                for file_line in context.lines:  # type: ignore
                    file_content += file_line
                found_secrets = super().analyze_line(
                    filename=filename, line=file_content, line_number=1,
                    context=context, raw_context=raw_context, **kwargs,
                )
                updated_secrets = self._get_updated_secrets(
                    found_secrets=found_secrets,
                    file_content=file_content,
                    split_by_newline=True,
                )
                output.update(updated_secrets)
                self._commit_hashes.add((filename, commit_hash))
            return output

        if filename not in self._analyzed_files \
                and 0 < self.get_file_size(filename) < PrivateKeyDetector.MAX_FILE_SIZE:
            self._analyzed_files.add(filename)
            file_content = self.read_file(filename)
            if file_content:
                found_secrets = super().analyze_line(
                    filename=filename, line=file_content, line_number=1,
                    context=context, raw_context=raw_context, **kwargs,
                )
                updated_secrets = self._get_updated_secrets(
                    found_secrets=found_secrets,
                    file_content=file_content,
                )
                output.update(updated_secrets)
        return output

    def _get_updated_secrets(
        self, found_secrets: Set[PotentialSecret],
        file_content: str,
        split_by_newline: Optional[bool] = False,
    ) -> Set[PotentialSecret]:
        updated_secrets: Set[PotentialSecret] = set()
        for sec in found_secrets:
            secret_val = sec.secret_value.strip() or ''  # type: ignore
            if split_by_newline and '\n' in secret_val:
                secret_val = secret_val.split('\n')[0]
            line_number = self.find_line_number(file_content, secret_val)
            updated_secrets.add(
                PotentialSecret(
                    type=self.secret_type,
                    filename=sec.filename,
                    secret=secret_val,
                    line_number=line_number,
                    is_verified=sec.is_verified,
                ),
            )
        return updated_secrets

    def analyze_string(self, string: str) -> Generator[str, None, None]:
        for regex in self.denylist:
            for match in regex.findall(string):
                if isinstance(match, tuple):
                    begin_key, secret_key, end_key = match
                    if begin_key:
                        yield secret_key if secret_key else begin_key
                else:
                    # only PuTTY-User-Key-File should not be a tuple
                    yield match

    def read_file(self, file_path: str) -> str:
        try:
            with open(file_path) as f:
                file_content = f.read()
                return file_content
        except Exception:
            return ''

    def get_file_size(self, file_path: str) -> int:
        try:
            return os.path.getsize(file_path)
        except Exception:
            return -1

    def find_line_number(
            self, file_content: str, substring: str, default_line_number: int = 1,
    ) -> int:
        if len(substring) == 0:
            return default_line_number
        lines = file_content.splitlines()

        for line_number, line in enumerate(lines, start=1):
            if substring in line:
                return line_number
        return default_line_number



import re

from .base import RegexBasedDetector


class ArtifactoryDetector(RegexBasedDetector):
    """Scans for Artifactory credentials."""
    secret_type = 'Artifactory Credentials'

    denylist = [
        # Artifactory tokens begin with AKC
        # API token:
        re.compile(r'(?:\s|=|:|"|^)AKC[a-zA-Z0-9]{10,200}(?:\s|"|$)'),
        # Artifactory encrypted passwords begin with AP[A-Z]
        # Keyword with Password:
        re.compile(r'(?i:artif|jfrog|buildkit)(?:.{0,100}\n?){0,2}(?:\s|=|:|"|^)(AP[\dABCDEF][a-zA-Z0-9]{8,200})(?:\s|"|$)'),
    ]



from abc import ABCMeta
from abc import abstractmethod
from typing import List

from ..custom_types import NamedIO


class BaseTransformer(metaclass=ABCMeta):
    """
    There are special filetypes (e.g. YAML) that work better with our line-based secrets parsing
    if we parse the file differently. In these cases, transformers can take the file, and parse
    it to meet the needs of the secret detector.

    While the transformation may not be an original copy, it just needs to proxy the original
    file so that we can obtain:
        1. The secret value
        2. The specific line that it's found on (for auditing purposes)
    """
    @property
    def is_eager(self) -> bool:
        """
        Eager transformers tend to be over-aggressive, and cause performance issues / false
        positives. We can make a transformer less eager through stricter validation checks
        on `should_parse_file`, however, in the cases where we are unable to do so, this flag
        informs the scanner to only use this transformer if all other methods fail to obtain
        secrets.
        """
        return False

    @abstractmethod
    def should_parse_file(self, filename: str) -> bool:
        raise NotImplementedError

    @abstractmethod
    def parse_file(self, file: NamedIO) -> List[str]:
        """
        :raises: ParsingError
        """
        raise NotImplementedError



import base64
import re
from collections import deque
from functools import lru_cache
from typing import Any
from typing import cast
from typing import Dict
from typing import Iterator
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Pattern
from typing import Tuple
from typing import Union

import yaml
from yaml.tokens import FlowEntryToken
from yaml.tokens import KeyToken

from ..custom_types import NamedIO
from ..util.filetype import determine_file_type
from ..util.filetype import FileType
from .base import BaseTransformer
from .exceptions import ParsingError


class YAMLTransformer(BaseTransformer):
    def should_parse_file(self, filename: str) -> bool:
        return determine_file_type(filename) == FileType.YAML

    def parse_file(self, file: NamedIO) -> List[str]:
        """
        :raises: ParsingError
        """
        try:
            items = sorted(YAMLFileParser(file), key=lambda x: x.line_number)
        except yaml.YAMLError:
            raise ParsingError

        lines: List[str] = []
        for item in items:
            while len(lines) < item.line_number - 1:
                lines.append('')

            value = item.value
            if isinstance(value, bytes):
                # binary strings in YAML are base64 encoded. https://yaml.org/type/binary.html
                # While the YAML parser already decodes it for us, we want to capture the *raw*
                # base64 encoded value for two reasons:
                #   1. Increases coverage
                #      Our Base64HighEntropyString plugin is tuned for base64 strings. Including
                #      other potential characters excludes it from this scan, with no real
                #      potential gain. While the entropy limit may be different, the fact that
                #      the string is processable is a win already.
                #
                #   2. Supports audit functionality
                #      When we convert this value to its unicode representation, we need to
                #      performs several hacks in order to be able to find the raw binary string
                #      again, during our audit process. Keeping it to its original value simplifies
                #      this process.
                value = base64.b64encode(value).decode()

            line = item.line.strip()
            # TODO: parse the difference between block_scalar styles, and handle appropriately.
            # See test cases for more details.

            comment = ''
            matches = _yaml_comment_regex().search(line)
            if matches:
                comment = matches.group(0)

            # YAML does the parsing for us, and the value we'll receive is a string proper
            # (rather than the raw value, like `ConfigFileTransformer`). We want to artificially
            # add quotes here since we know they are strings, HighEntropyString will benefit
            # from this, and all other plugins don't care.
            #
            # However, if there is a quote inside, we need to escape it.
            value = value.replace('"', '\\"')

            lines.append(f'{item.key}: "{value}"{comment}')     # type: ignore

        return lines


@lru_cache(maxsize=1)
def _yaml_comment_regex() -> Pattern:
    """
    From https://yaml-multiline.info/, it states that `#` cannot appear *after* a space
    or a newline, otherwise it will be a syntax error (for multiline strings that don't
    use a block scalar). This applies to single lines as well: for example, `a#b` will be
    treated as a single value, but `a #b` will only capture `a`, leaving `#b` as a comment.

    For lines that *do* use a block scalar, the YAML parser will throw a syntax error if
    there is additional text on the same line as the block scalar. Comments however, are fine.
    e.g.
        key: | # this is ok
        blah

        key: | but this is not
        blah

    Given that we've made it to this stage, we can assume the YAML file is syntactically
    correct. Therefore, if we add whitespace before the comment character, we can know that
    everything else *after* the comment character is a comment for a given line.
    """
    return re.compile(r'(\s+#[\S ]*)')


class YAMLValue(NamedTuple):
    key: str
    value: Union[str, bytes]
    line_number: int
    line: str


class YAMLFileParser:
    """
    Yaml config files are interesting, because they don't necessarily conform
    to our basic regex for detecting HighEntropyStrings as strings don't
    need to be quoted.

    This causes interesting issues, because our regex won't catch non-quoted
    strings, and if we ignore the quoting requirement, then we increase our
    false positive rate, because any long string would have high entropy.

    Therefore, we take a different approach: intercept the parsing of the yaml
    file to identify string values. This assumes:

        1. Secrets are strings or binaries
        2. Secrets are not keys

    Then, we calculate the entropy of those string values.

    The difficulty comes from determining the line number which these values
    come from. To do this, we transform the string into a dictionary of
    meta-tags, in the following format:

    >>> {
        'key': {
            '__value__': value,
            '__line__': <line_number>,
        }
    }

    This way, we can quickly identify the line number for auditing at a later
    stage.

    This parsing method is inspired by https://stackoverflow.com/a/13319530.
    """

    def __init__(self, file: NamedIO):
        self.content = file.read()

        self.loader = yaml.SafeLoader(self.content)
        self.loader.compose_node = self._compose_node_shim  # type: ignore

        self.is_inline_flow_mapping_key = False
        self.loader.parse_flow_mapping_key = self._parse_flow_mapping_key_shim  # type: ignore

    def json(self) -> Dict[str, Any]:
        return cast(Dict[str, Any], self.loader.get_single_data())

    def __iter__(self) -> Iterator[YAMLValue]:
        """
        :returns: (value, line_number)
        """
        # Used to obtain inline comments
        lines = self.content.splitlines()

        to_search = deque([self.json()])
        while to_search:
            item: Any = to_search.popleft()

            if not item:
                # mainly for base case (e.g. if file is all comments)
                continue

            # If it doesn't have our meta-tags, it's not a value worth scanning.
            try:
                if '__line__' not in item:
                    if isinstance(item, str):
                        continue

                    # However, we need to recursively search in the dictionary for other such values
                    # that we may care about.
                    try:
                        to_search.extend(item.values())
                    except AttributeError:
                        # This is an array
                        to_search.extend(item)

                    continue
            except TypeError:
                # e.g. if item is a float.
                continue

            yield YAMLValue(
                key=item['__original_key__'],
                value=item['__value__'],
                line_number=item['__line__'],

                # We extract this separately because the parser drops the comments
                # (at least up to version 3.13).
                # https://github.com/yaml/pyyaml/blob/a2d481b8dbd2b352cb001f07091ccf669227290f/lib3/yaml/scanner.py#L749
                # The line value feeds into the filters, and helps us tune false positives.
                line=lines[item['__line__'] - 1],
            )

    def _compose_node_shim(
        self,
        parent: Optional[yaml.nodes.Node],
        index: Optional[yaml.nodes.Node],
    ) -> Optional[yaml.nodes.Node]:
        line = (
            self.loader.marks[-1].line
            if self.is_inline_flow_mapping_key
            else self.loader.line
        )

        node = yaml.composer.Composer.compose_node(self.loader, parent, index)  # type: ignore
        if node is None:
            return None

        node.__line__ = line + 1    # type: ignore

        if node.tag.endswith(':map'):
            # Reset the inline flow mapping key when the end of a mapping is reached
            # to avoid complications with empty mappings
            self.is_inline_flow_mapping_key = False
            return _tag_dict_values(cast(yaml.nodes.MappingNode, node))

        # TODO: Not sure if need to do :seq

        return node

    def _parse_flow_mapping_key_shim(
        self,
        first: bool = False,
    ) -> yaml.nodes.Node:
        # There exists an edge case when a key and flow mapping start character `{` are on the same
        # line (Ex. '{key: value}) followed by an empty line. The parser will produce an off-by-one
        # error for the line number that it tracks internally. Since we track the start of the
        # mapping, we will use this line number when we are processing:
        # A) The first key in an inline dictionary where the flow mapping start character is on the
        # same line as the key
        # B) The n key of an inline dictionary that is followed by a FlowEntryToken (',') and
        # KeyToken ('key:')
        is_inline_dictionary = (
            first
            and self.loader.marks[-1].line == self.loader.peek_token().start_mark.line  # type:ignore[no-untyped-call]
            or self._check_next_tokens_shim(FlowEntryToken, KeyToken)
        )

        if is_inline_dictionary:
            self.is_inline_flow_mapping_key = True
        else:
            self.is_inline_flow_mapping_key = False

        return cast(
            'yaml.nodes.Node',
            yaml.parser.Parser.parse_flow_mapping_key(self.loader, first),  # type:ignore[no-untyped-call]
        )

    def _check_next_tokens_shim(
        self,
        *choices: Any,
    ) -> bool:
        """Check the next tokens type match the argument list of token types."""
        result = True
        i = 0

        if self.loader.tokens:
            if not choices:
                return result
            for choice in choices:
                if i < len(self.loader.tokens):
                    result = result and isinstance(self.loader.tokens[i], choice)
                    i += 1
        else:
            result = False

        return result


def _tag_dict_values(map_node: yaml.nodes.MappingNode) -> yaml.nodes.MappingNode:
    """
    :param map_node: It looks like map_node.value contains a list of
        pair tuples, corresponding to key,value pairs.
    """
    new_values = []
    for key, value in map_node.value:
        if not (
            value.tag.endswith(':str') or
            value.tag.endswith(':binary')
        ):
            new_values.append((key, value))
            continue

        augmented_string = yaml.nodes.MappingNode(
            tag=map_node.tag,
            value=[
                _create_key_value_pair_for_mapping_node_value(
                    key='__value__',
                    value=value.value,
                    tag=value.tag,
                ),
                _create_key_value_pair_for_mapping_node_value(
                    key='__line__',
                    value=str(value.__line__),
                    tag='tag:yaml.org,2002:int',
                ),
                _create_key_value_pair_for_mapping_node_value(
                    key='__original_key__',
                    value=key.value,
                    tag='tag:yaml.org,2002:str',
                ),
            ],
        )

        new_values.append((key, augmented_string))

    output = yaml.nodes.MappingNode(
        tag=map_node.tag,
        value=new_values,
        start_mark=map_node.start_mark,
        end_mark=map_node.end_mark,
        flow_style=map_node.flow_style,
    )
    return output


def _create_key_value_pair_for_mapping_node_value(
    key: str,
    value: Any,
    tag: str,
) -> Tuple[yaml.nodes.ScalarNode, yaml.nodes.ScalarNode]:
    return (
        yaml.nodes.ScalarNode(
            tag='tag:yaml.org,2002:str',
            value=key,
        ),
        yaml.nodes.ScalarNode(
            tag=tag,
            value=value,
        ),
    )



import inspect
from typing import Any
from typing import Callable
from typing import cast
from typing import Tuple
from typing import Union

from ..custom_types import SelfAwareCallable


def call_function_with_arguments(
    func: Union[Callable, SelfAwareCallable],
    **kwargs: Any,
) -> Any:
    """
    :raises: TypeError
    """
    # First, we ensure that the function we're going to inject values into is self-aware.
    function = func if isinstance(func, SelfAwareCallable) else make_function_self_aware(func)

    # If `function` is derived from a method, we add the instance of the class by default.
    # However, if `function` is a method itself, it will already carry the reference of the
    # instance, so we don't need to explicitly include it.
    if inspect.ismethod(func) and not inspect.ismethod(function):
        # We also use get_injectable_variables (instead of hardcoding "self") to make sure that
        # this also handles cases where the developer doesn't name the first parameter `self`.
        kwargs[get_injectable_variables(func)[0]] = func.__self__

    variables_to_inject = set(kwargs.keys())
    values = {
        key: kwargs[key]
        for key in (variables_to_inject & function.injectable_variables)
    }

    return function(**values)


def make_function_self_aware(func: Callable) -> SelfAwareCallable:
    """
    A SelfAwareCallable is one that is aware of its own injectable variables, through the
    `func.injectable_variables` attribute.
    """
    if hasattr(func, 'injectable_variables'):
        return cast(SelfAwareCallable, func)

    # We can't add arbitrary attributes to methods, but we can to functions. Therefore,
    # we need to reference the underlying function itself.
    if inspect.ismethod(func):
        klass = func.__self__.__class__
        function = getattr(klass, func.__name__)
        function.injectable_variables = set(get_injectable_variables(func))

        function.path = f'{klass}.{func.__name__}'
    else:
        function = func
        function.path = func.__name__

    return cast(SelfAwareCallable, function)


def get_injectable_variables(func: Callable) -> Tuple[str, ...]:
    """
    The easiest way to understand this is to see it as an example:
        >>> def func(a, b=1, *args, c, d=2, **kwargs):
        ...     e = 5
        >>>
        >>> print(func.__code__.co_varnames)
        ('a', 'b', 'c', 'd', 'args', 'kwargs', 'e')
        >>> print(func.__code__.co_argcount)    # `a` and `b`
        2
        >>> print(func.__code__.co_kwonlyargcount)  # `c` and `d`
        2
    """
    variable_names = func.__code__.co_varnames
    arg_count = func.__code__.co_argcount + func.__code__.co_kwonlyargcount

    return variable_names[:arg_count]



from typing import Any
from typing import Dict


def upgrade(baseline: Dict[str, Any]) -> None:
    for function in [
        _add_new_default_filters,
    ]:
        function(baseline)


def _add_new_default_filters(baseline: Dict[str, Any]) -> None:
    baseline['filters_used'].extend([
        {
            'path': 'detect_secrets.filters.heuristic.is_lock_file',
        },
        {
            'path': 'detect_secrets.filters.heuristic.is_not_alphanumeric_string',
        },
        {
            'path': 'detect_secrets.filters.heuristic.is_swagger_file',
        },
    ])



from collections.abc import Iterator
from typing import Any
from typing import Sequence

from . import io
from ..core.secrets_collection import SecretsCollection


def get_secret_iterator(baseline: SecretsCollection) -> 'BidirectionalIterator':
    """
    :returns: (index, filename, secret)
    """
    unlabelled_secrets = [
        secret
        for _, secret in baseline
        if secret.is_secret is None
    ]

    if not unlabelled_secrets:
        io.print_message('Nothing to audit!')
        return BidirectionalIterator([])

    return BidirectionalIterator(unlabelled_secrets)


class BidirectionalIterator(Iterator):
    def __init__(self, collection: Sequence):
        self.collection = collection
        self.index = -1  # Starts on -1, as index is increased _before_ getting result
        self.step_back_once = False

    def __next__(self) -> Any:
        if self.step_back_once:
            self.index -= 1
            self.step_back_once = False
        else:
            self.index += 1

        if self.index < 0:
            raise StopIteration

        try:
            result = self.collection[self.index]
        except IndexError:
            raise StopIteration

        return result

    def next(self) -> Any:  # pragma: no cover
        return self.__next__()

    def step_back_on_next_iteration(self) -> None:
        self.step_back_once = True

    def can_step_back(self) -> bool:
        return self.index > 0

    def __iter__(self) -> 'BidirectionalIterator':  # pragma: no cover
        return self



import argparse
from typing import cast


def add_audit_action(parent: argparse._SubParsersAction) -> argparse.ArgumentParser:
    parser = parent.add_parser(
        'audit',
        help='Manually assesses a baseline to determine validity of secrets found.',
        description=(
            'Auditing a baseline allows analysts to label results, and optimize plugins for '
            'the highest signal-to-noise ratio for their environment.'
        ),
    )

    parser.add_argument(
        'filename',
        nargs='+',
        help=(
            'Audit a given baseline file to distinguish the difference '
            'between false and true positives.'
        ),
    )

    _add_mode_parser(parser)
    _add_report_module(parser)
    _add_statistics_module(parser)
    return cast(argparse.ArgumentParser, parser)


def _add_mode_parser(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        '--diff',
        action='store_true',
        help=(
            'Allows the comparison of two baseline files, in order to '
            'effectively distinguish the difference between various '
            'plugin configurations.'
        ),
    )

    parser.add_argument(
        '--stats',
        action='store_true',
        help=(
            'Displays the results of an interactive auditing session '
            'which have been saved to a baseline file.'
        ),
    )

    parser.add_argument(
        '--report',
        action='store_true',
        help=(
            'Displays a report with the secrets detected'
        ),
    )


def _add_report_module(parent: argparse.ArgumentParser) -> None:
    parser = parent.add_argument_group(
        title='reporting',
        description=(
            'Display a report with all the findings and the made decisions. '
            'To be used with the report mode (--report).'
        ),
    )

    report_parser = parser.add_mutually_exclusive_group()
    report_parser.add_argument(
        '--only-real',
        action='store_true',
        help=(
            'Only includes real secrets in the report'
        ),
    )

    report_parser.add_argument(
        '--only-false',
        action='store_true',
        help=(
            'Only includes false positives in the report'
        ),
    )


def _add_statistics_module(parent: argparse.ArgumentParser) -> None:
    parser = parent.add_argument_group(
        title='analytics',
        description=(
            'Quantify the success of your plugins based on the labelled results '
            'in your baseline. To be used with the statistics mode (--stats).'
        ),
    )

    parser.add_argument(
        '--json',
        action='store_true',
        help='Outputs results in a machine-readable format.',
    )


def parse_args(args: argparse.Namespace) -> None:
    if args.action != 'audit':
        return

    if args.diff and len(args.filename) != 2:
        raise argparse.ArgumentTypeError(
            '--diff mode requires two files to compare with each other.',
        )
    elif not args.diff and len(args.filename) != 1:
        raise argparse.ArgumentTypeError(
            'Can only specify one baseline at a time.',
        )



import argparse
import sys
from typing import Callable
from typing import List
from typing import Optional

from . import audit
from . import baseline
from . import filters
from . import plugins
from . import scan
from ...settings import get_settings
from .common import initialize_plugin_settings
from detect_secrets.__version__ import VERSION


class ParserBuilder:
    def __init__(self) -> None:
        self._parser = argparse.ArgumentParser()
        self._post_processors: List[Callable[[argparse.Namespace], None]] = [
            initialize_plugin_settings,
        ]

        self.add_default_options()

    def add_default_options(self) -> 'ParserBuilder':
        self._parser.add_argument(
            '-v',
            '--verbose',
            action='count',
            help='Verbose mode.',
        )
        self._parser.add_argument(
            '--version',
            action='version',
            version=VERSION,
            help='Display version information.',
        )
        self._parser.add_argument(
            '-C',
            metavar='<path>',
            dest='custom_root',
            nargs=1,
            default=[''],
            help=(
                'Run as if detect-secrets was started in <path>, rather than in the current '
                'working directory.'
            ),
        )
        self._parser.add_argument(
            '-c',
            '--cores',
            dest='num_cores',
            nargs=1,
            type=int,
            default=[None],
            help=(
                'Specify the number of cores to use for parallel processing. Defaults to '
                'using the max cores on the current host.'
            ),
        )
        return self

    def add_console_use_arguments(self) -> 'ParserBuilder':
        subparser = self._parser.add_subparsers(dest='action')
        self._post_processors = [_assert_action_is_specified, *self._post_processors]

        parser = scan.add_scan_action(subparser)

        # NOTE: This ordering is important.
        #   1. Baselines will be handled accordingly, and the global settings object will
        #      be initialized with a certain state.
        #   2. Scan options can override this (e.g. --force-use-all-plugins)
        #   3. Plugin options can override this again (e.g. disabling plugins, or different configs)
        #
        # In a similar way, the filter options must come after the settings object is initialized.
        self._post_processors.append(scan.parse_args)
        self.add_plugin_options(parser, action_filter='scan')
        self.add_filter_options(parser, action_filter='scan')

        # NOTE: scan allows a baseline, but we need to override the first post_processor
        self._post_processors[1] = baseline.parse_args

        audit.add_audit_action(subparser)
        self._post_processors.append(audit.parse_args)
        return self

    def add_pre_commit_arguments(self) -> 'ParserBuilder':
        self._parser.add_argument(
            'filenames',
            nargs='*',
            help='Filenames to check.',
        )
        self._parser.add_argument(
            '--json',
            action='store_true',
            help='Print detect-secrets-hook output as JSON',
        )
        self.add_baseline_options(
            help=(
                'Explicitly ignore secrets through a baseline generated by `detect-secrets scan`'
            ),
        )
        self.add_plugin_options()
        self.add_filter_options()

        return self

    def add_baseline_options(self, help: str = '') -> 'ParserBuilder':
        baseline.add_baseline_option(self._parser, help=help)
        for index, processor in enumerate(self._post_processors):
            if processor == initialize_plugin_settings:
                self._post_processors[index] = baseline.parse_args
                break

        return self

    def add_plugin_options(
        self,
        parser: Optional[argparse.ArgumentParser] = None,
        action_filter: Optional[str] = None,
    ) -> 'ParserBuilder':
        if not parser:
            parser = self._parser

        plugins.add_plugin_options(parser)

        if action_filter:
            self._post_processors.append(
                _action_specific_post_processor(action_filter, plugins.parse_args),
            )
        else:
            self._post_processors.append(plugins.parse_args)

        return self

    def add_filter_options(
        self,
        parser: Optional[argparse.ArgumentParser] = None,
        action_filter: Optional[str] = None,
    ) -> 'ParserBuilder':
        if not parser:
            parser = self._parser

        filters.add_filter_options(parser)

        if action_filter:
            self._post_processors.append(
                _action_specific_post_processor(action_filter, filters.parse_args),
            )
        else:
            self._post_processors.append(filters.parse_args)

        return self

    def parse_args(self, argv: Optional[List[str]] = None) -> argparse.Namespace:
        args = self._parser.parse_args(argv)
        try:
            for processor in self._post_processors:
                processor(args)
        except argparse.ArgumentTypeError:
            # TODO: Better help text?
            self._parser.print_usage(sys.stderr)
            sys.exit(1)

        args.custom_root = args.custom_root[0]
        if args.custom_root:
            # This filter assumes current working directory, which will fail if we're running
            # from a different directory.
            # TODO: Maybe adjust this so that it is directory agnostic?
            get_settings().disable_filters('detect_secrets.filters.common.is_invalid_file')

            # Abide by the Principle of Least Surprise, and have the default value be the
            # custom root directory itself.
            if args.path == ['.']:
                args.path = [args.custom_root]

        args.num_cores = args.num_cores[0]

        return args


def _assert_action_is_specified(args: argparse.Namespace) -> None:
    if not args.action:
        raise argparse.ArgumentTypeError('Unspecified action.')


def _action_specific_post_processor(action: str, processor: Callable) -> Callable:
    def wrapped(args: argparse.Namespace) -> None:
        if args.action != action:
            return

        processor(args)

    return wrapped



"""
This plugin searches for Slack tokens
"""
import re
from typing import Any
from typing import cast
from typing import Dict

import requests

from ..constants import VerifiedResult
from .base import RegexBasedDetector


class SlackDetector(RegexBasedDetector):
    """Scans for Slack tokens."""
    secret_type = 'Slack Token'

    denylist = (
        # Slack Token
        re.compile(r'xox(?:a|b|p|o|s|r)-(?:\d+-)+[a-z0-9]+', flags=re.IGNORECASE),
        # Slack Webhooks
        re.compile(
            r'https://hooks\.slack\.com/services/T[a-zA-Z0-9_]+/B[a-zA-Z0-9_]+/[a-zA-Z0-9_]+',
            flags=re.IGNORECASE | re.VERBOSE,
        ),
    )

    def verify(self, secret: str) -> VerifiedResult:  # pragma: no cover
        if secret.startswith('https://hooks.slack.com/services/T'):
            response = requests.post(
                secret,
                json={
                    'text': '',
                },
            )
            valid = response.text in ['missing_text_or_fallback_or_attachments', 'no_text']
        else:
            response = requests.post(
                'https://slack.com/api/auth.test',
                data={
                    'token': secret,
                },
            ).json()
            valid = cast(Dict[str, Any], response)['ok']

        return (
            VerifiedResult.VERIFIED_TRUE
            if valid
            else VerifiedResult.VERIFIED_FALSE
        )
